@@ -1,20 +1,22 @@
 import asyncio
+import copy
 import enum
 import ipaddress
 import logging
 import random
 import secrets
 import socket
+import threading
 from itertools import count
 from typing import Dict, List, Optional, Set, Text, Tuple, Union, cast
 
 import netifaces
 
-from . import stun, turn
+from . import mdns, stun, turn
 from .candidate import Candidate, candidate_foundation, candidate_priority
 from .utils import random_string
 
-logger = logging.getLogger("ice")
+logger = logging.getLogger(__name__)
 
 ICE_COMPLETED = 1
 ICE_FAILED = 2
@@ -25,6 +27,29 @@ CONSENT_INTERVAL = 5
 connection_id = count()
 protocol_id = count()
 
+_mdns = threading.local()
+
+
+async def get_or_create_mdns_protocol(subscriber: object) -> mdns.MDnsProtocol:
+    if not hasattr(_mdns, "lock"):
+        _mdns.lock = asyncio.Lock()
+        _mdns.protocol = None
+        _mdns.subscribers = set()
+    async with _mdns.lock:
+        if _mdns.protocol is None:
+            _mdns.protocol = await mdns.create_mdns_protocol()
+        _mdns.subscribers.add(subscriber)
+    return _mdns.protocol
+
+
+async def unref_mdns_protocol(subscriber: object) -> None:
+    if hasattr(_mdns, "lock"):
+        async with _mdns.lock:
+            _mdns.subscribers.discard(subscriber)
+            if _mdns.protocol and not _mdns.subscribers:
+                await _mdns.protocol.close()
+                _mdns.protocol = None
+
 
 def candidate_pair_priority(
     local: Candidate, remote: Candidate, ice_controlling: bool
@@ -102,8 +127,6 @@ def sort_candidate_pairs(pairs, ice_controlling: bool) -> None:
 def validate_remote_candidate(candidate: Candidate) -> Candidate:
     """
     Check the remote candidate is supported.
-
-    mDNS candidates are not supported yet.
     """
     if candidate.type not in ["host", "relay", "srflx"]:
         raise ValueError('Unexpected candidate type "%s"' % candidate.type)
@@ -158,8 +181,9 @@ class StunProtocol(asyncio.DatagramProtocol):
 
     def connection_lost(self, exc: Exception) -> None:
         self.__log_debug("connection_lost(%s)", exc)
-        self.receiver.data_received(None, self.local_candidate.component)
-        self.__closed.set_result(True)
+        if not self.__closed.done():
+            self.receiver.data_received(None, None)
+            self.__closed.set_result(True)
 
     def connection_made(self, transport) -> None:
         self.__log_debug("connection_made(%s)", transport)
@@ -209,7 +233,6 @@ class StunProtocol(asyncio.DatagramProtocol):
 
         if integrity_key is not None:
             request.add_message_integrity(integrity_key)
-            request.add_fingerprint()
 
         transaction = stun.Transaction(
             request, addr, self, retransmissions=retransmissions
@@ -237,6 +260,14 @@ class StunProtocol(asyncio.DatagramProtocol):
         return "protocol(%s)" % self.id
 
 
+class ConnectionEvent:
+    pass
+
+
+class ConnectionClosed(ConnectionEvent):
+    pass
+
+
 class Connection:
     """
     An ICE connection for a single media stream.
@@ -286,6 +317,7 @@ class Connection:
         self.turn_transport = turn_transport
 
         # private
+        self._closed = False
         self._components = set(range(1, components + 1))
         self._check_list: List[CandidatePair] = []
         self._check_list_done = False
@@ -293,6 +325,7 @@ class Connection:
         self._early_checks: List[
             Tuple[stun.Message, Tuple[str, int], StunProtocol]
         ] = []
+        self._event_waiter: Optional[asyncio.Future[ConnectionEvent]] = None
         self._id = next(connection_id)
         self._local_candidates: List[Candidate] = []
         self._local_candidates_end = False
@@ -318,33 +351,12 @@ class Connection:
     @property
     def remote_candidates(self) -> List[Candidate]:
         """
-        Remote candidates, which you need to set.
-
-        Assigning this attribute will automatically signal end-of-candidates.
-        If you will be adding more remote candidates in the future, use the
-        :meth:`add_remote_candidate` method instead.
+        Remote candidates, which you need to populate using
+        :meth:`add_remote_candidate`.
         """
         return self._remote_candidates[:]
 
-    @remote_candidates.setter
-    def remote_candidates(self, value: List[Candidate]) -> None:
-        if self._remote_candidates_end:
-            raise ValueError("Cannot set remote candidates after end-of-candidates.")
-
-        # validate the remote candidates
-        self._remote_candidates = []
-        for remote_candidate in value:
-            try:
-                validate_remote_candidate(remote_candidate)
-            except ValueError:
-                continue
-            self._remote_candidates.append(remote_candidate)
-
-        # end-of-candidates
-        self._prune_components()
-        self._remote_candidates_end = True
-
-    def add_remote_candidate(self, remote_candidate: Candidate) -> None:
+    async def add_remote_candidate(self, remote_candidate: Candidate) -> None:
         """
         Add a remote candidate or signal end-of-candidates.
 
@@ -361,10 +373,31 @@ class Connection:
             self._remote_candidates_end = True
             return
 
+        # resolve mDNS candidate
+        if mdns.is_mdns_hostname(remote_candidate.host):
+            mdns_protocol = await get_or_create_mdns_protocol(self)
+            remote_addr = await mdns_protocol.resolve(remote_candidate.host)
+            if remote_addr is None:
+                self.__log_info(
+                    f'Remote candidate "{remote_candidate.host}" could not be resolved'
+                )
+                return
+            self.__log_info(
+                f'Remote candidate "{remote_candidate.host}" resolved to {remote_addr}'
+            )
+
+            copy_candidate = copy.copy(remote_candidate)
+            copy_candidate.host = remote_addr
+            await self.add_remote_candidate(copy_candidate)
+            return
+
         # validate the remote candidate
         try:
             validate_remote_candidate(remote_candidate)
-        except ValueError:
+        except ValueError as e:
+            self.__log_info(
+                f'Remote candidate "{remote_candidate.host}" is not valid: {e}'
+            )
             return
         self._remote_candidates.append(remote_candidate)
 
@@ -388,10 +421,12 @@ class Connection:
             addresses = get_host_addresses(
                 use_ipv4=self._use_ipv4, use_ipv6=self._use_ipv6
             )
-            for component in self._components:
-                self._local_candidates += await self.get_component_candidates(
-                    component=component, addresses=addresses
-                )
+            coros = [
+                self.get_component_candidates(component=component, addresses=addresses)
+                for component in self._components
+            ]
+            for candidates in await asyncio.gather(*coros):
+                self._local_candidates += candidates
             self._local_candidates_end = True
 
     def get_default_candidate(self, component: int) -> Optional[Candidate]:
@@ -474,12 +509,34 @@ class Connection:
         if self._check_list and not self._check_list_done:
             await self._check_list_state.put(ICE_FAILED)
 
+        # unreference mDNS
+        await unref_mdns_protocol(self)
+
         self._nominated.clear()
         for protocol in self._protocols:
             await protocol.close()
         self._protocols.clear()
         self._local_candidates.clear()
 
+        # emit event
+        if not self._closed:
+            self._emit_event(ConnectionClosed())
+            self._closed = True
+
+    async def get_event(self) -> Optional[ConnectionEvent]:
+        """
+        Return the next `ConnectionEvent` or `None` if the connection is
+        already closed.
+
+        This method may only be called once at a time.
+        """
+        assert self._event_waiter is None, "already awaiting event"
+        if self._closed:
+            return None
+        loop = asyncio.get_event_loop()
+        self._event_waiter = loop.create_future()
+        return await asyncio.shield(self._event_waiter)
+
     async def recv(self) -> bytes:
         """
         Receive the next datagram.
@@ -769,6 +826,12 @@ class Connection:
         self.__log_info("Check %s %s -> %s", pair, pair.state, state)
         pair.state = state
 
+    def _emit_event(self, event: ConnectionEvent) -> None:
+        if self._event_waiter is not None:
+            waiter = self._event_waiter
+            self._event_waiter = None
+            waiter.set_result(event)
+
     def _find_pair(
         self, protocol: StunProtocol, remote_candidate: Candidate
     ) -> Optional[CandidatePair]:
@@ -784,8 +847,10 @@ class Connection:
         self, component: int, addresses: List[str], timeout: int = 5
     ) -> List[Candidate]:
         candidates = []
-
         loop = asyncio.get_event_loop()
+
+        # gather host candidates
+        host_protocols = []
         for address in addresses:
             # create transport
             try:
@@ -796,7 +861,7 @@ class Connection:
                 self.__log_info("Could not bind to %s - %s", address, exc)
                 continue
             protocol = cast(StunProtocol, protocol)
-            self._protocols.append(protocol)
+            host_protocols.append(protocol)
 
             # add host candidate
             candidate_address = protocol.transport.get_extra_info("sockname")
@@ -810,15 +875,20 @@ class Connection:
                 type="host",
             )
             candidates.append(protocol.local_candidate)
+        self._protocols += host_protocols
 
         # query STUN server for server-reflexive candidates (IPv4 only)
         if self.stun_server:
-            fs = []
-            for protocol in self._protocols:
+            tasks = []
+            for protocol in host_protocols:
                 if ipaddress.ip_address(protocol.local_candidate.host).version == 4:
-                    fs.append(server_reflexive_candidate(protocol, self.stun_server))
-            if len(fs):
-                done, pending = await asyncio.wait(fs, timeout=timeout)
+                    tasks.append(
+                        asyncio.ensure_future(
+                            server_reflexive_candidate(protocol, self.stun_server)
+                        )
+                    )
+            if len(tasks):
+                done, pending = await asyncio.wait(tasks, timeout=timeout)
                 candidates += [
                     task.result() for task in done if task.exception() is None
                 ]
@@ -946,7 +1016,6 @@ class Connection:
         )
         response.attributes["XOR-MAPPED-ADDRESS"] = addr
         response.add_message_integrity(self.local_password.encode("utf8"))
-        response.add_fingerprint()
         protocol.send_stun(response, addr)
 
         if not self._check_list:
@@ -968,7 +1037,6 @@ class Connection:
         )
         response.attributes["ERROR-CODE"] = error_code
         response.add_message_integrity(self.local_password.encode("utf8"))
-        response.add_fingerprint()
         protocol.send_stun(response, addr)
 
     def sort_check_list(self) -> None:
