@@ -1,13 +1,17 @@
 import argparse
 import asyncio
+import contextlib
 import logging
 import os
 import ssl
 import struct
 import time
+from typing import Optional, Tuple
 
 from aioice import stun
+from aioice.ice import get_host_addresses
 from aioice.turn import (
+    DEFAULT_ALLOCATION_LIFETIME,
     UDP_TRANSPORT,
     TurnStreamMixin,
     is_channel_data,
@@ -24,7 +28,7 @@ CERT_FILE = os.path.join(ROOT, "turnserver.crt")
 KEY_FILE = os.path.join(ROOT, "turnserver.key")
 
 
-def create_self_signed_cert(name):
+def create_self_signed_cert(name="localhost"):
     from OpenSSL import crypto
 
     # create key pair
@@ -106,6 +110,13 @@ class TurnServerMixin:
             self.send_stun(response, addr)
             return
 
+        # generate failure for test purposes
+        if self.server.simulated_failure:
+            response = self.error_response(message, *self.server.simulated_failure)
+            self.server.simulated_failure = None
+            self.send_stun(response, addr)
+            return
+
         if "USERNAME" not in message.attributes:
             response = self.error_response(message, 401, "Unauthorized")
             response.attributes["NONCE"] = random_string(16).encode("ascii")
@@ -135,7 +146,6 @@ class TurnServerMixin:
             )
 
         response.add_message_integrity(integrity_key)
-        response.add_fingerprint()
         self.send_stun(response, addr)
 
     async def handle_allocate(self, message, addr, integrity_key):
@@ -151,7 +161,7 @@ class TurnServerMixin:
                 message, 442, "Unsupported transport protocol"
             )
         else:
-            lifetime = message.attributes.get("LIFETIME", self.server.default_lifetime)
+            lifetime = message.attributes.get("LIFETIME", DEFAULT_ALLOCATION_LIFETIME)
             lifetime = min(lifetime, self.server.maximum_lifetime)
 
             # create allocation
@@ -181,7 +191,6 @@ class TurnServerMixin:
 
         # send response
         response.add_message_integrity(integrity_key)
-        response.add_fingerprint()
         self.send_stun(response, addr)
 
     def handle_binding(self, message, addr):
@@ -255,7 +264,7 @@ class TurnServerMixin:
             allocation.expiry = time.time() + lifetime
         else:
             logger.info("Allocation deleted %s", allocation.relayed_address)
-            del self.server.allocations[key]
+            self.server._remove_allocation(key)
 
         # build response
         response = stun.Message(
@@ -300,25 +309,32 @@ class TurnServer:
 
     def __init__(self, realm="test", users={}):
         self.allocations = {}
-        self.default_lifetime = 600
         self.maximum_lifetime = 3600
         self.realm = realm
+        self.simulated_failure: Optional[Tuple[int, str]] = None
         self.users = users
 
         self._expire_handle = None
 
     async def close(self):
-        # start expiry loop
+        # stop expiry loop
         self._expire_handle.cancel()
 
+        # close allocations
+        for key in list(self.allocations.keys()):
+            self._remove_allocation(key)
+
+        # shutdown servers
         self.tcp_server.close()
+        self.tls_server.close()
         self.udp_server.transport.close()
-        await self.tcp_server.wait_closed()
+        await asyncio.gather(
+            self.tcp_server.wait_closed(), self.tls_server.wait_closed()
+        )
 
     async def listen(self, port=0, tls_port=0):
         loop = asyncio.get_event_loop()
-        hostaddr = "127.0.0.1"
-        hostname = "localhost"
+        hostaddr = get_host_addresses(use_ipv4=True, use_ipv6=False)[0]
 
         # listen for TCP
         self.tcp_server = await loop.create_server(
@@ -337,14 +353,13 @@ class TurnServer:
         # listen for TLS
         ssl_context = ssl.SSLContext()
         ssl_context.load_cert_chain(CERT_FILE, KEY_FILE)
-        # create_self_signed_cert(hostname)
         self.tls_server = await loop.create_server(
             lambda: TurnServerTcpProtocol(server=self),
             host=hostaddr,
             port=tls_port,
             ssl=ssl_context,
         )
-        self.tls_address = (hostname, self.tls_server.sockets[0].getsockname()[1])
+        self.tls_address = self.tls_server.sockets[0].getsockname()
         logger.info("Listening for TLS on %s", self.tls_address)
 
         # start expiry loop
@@ -353,13 +368,27 @@ class TurnServer:
     async def _expire_allocations(self):
         while True:
             now = time.time()
-            for key, allocation in self.allocations.items():
+            for key, allocation in list(self.allocations.items()):
                 if allocation.expiry < now:
                     logger.info("Allocation expired %s", allocation.relayed_address)
-                    del self.allocations[key]
+                    self.server._remove_allocation(key)
 
             await asyncio.sleep(1)
 
+    def _remove_allocation(self, key):
+        allocation = self.allocations.pop(key)
+        allocation.transport.close()
+
+
+@contextlib.asynccontextmanager
+async def run_turn_server(**kwargs):
+    server = TurnServer(**kwargs)
+    await server.listen()
+    try:
+        yield server
+    finally:
+        await server.close()
+
 
 if __name__ == "__main__":
     parser = argparse.ArgumentParser(description="STUN / TURN server")
